use std::{
    collections::HashMap,
    ops::{Deref, DerefMut},
    str::FromStr,
    sync::Arc,
    time::Duration,
};

use chrono::{DateTime, Utc};
use jabber::{connection::Tls, jabber_stream::bound_stream::BoundJabberReader};
use stanza::client::Stanza;
use tokio::{
    sync::{mpsc, oneshot, Mutex},
    task::{JoinHandle, JoinSet},
};
use tracing::info;
use uuid::Uuid;

use crate::{
    chat::{Body, Message},
    db::Db,
    error::{Error, IqError, MessageRecvError, PresenceError, ReadError, RosterError},
    presence::{Offline, Online, Presence, Show},
    roster::Contact,
    UpdateMessage,
};

use super::{write::WriteHandle, SupervisorCommand};

pub struct Read {
    control_receiver: mpsc::Receiver<ReadControl>,
    stream: BoundJabberReader<Tls>,
    on_crash: oneshot::Sender<(
        Db,
        mpsc::Sender<UpdateMessage>,
        JoinSet<()>,
        mpsc::Sender<SupervisorCommand>,
        WriteHandle,
        Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
    )>,
    db: Db,
    update_sender: mpsc::Sender<UpdateMessage>,
    supervisor_control: mpsc::Sender<SupervisorCommand>,
    write_handle: WriteHandle,
    tasks: JoinSet<()>,
    disconnecting: bool,
    disconnect_timedout: oneshot::Receiver<()>,
    // TODO: use proper stanza ids
    pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
}

impl Read {
    fn new(
        control_receiver: mpsc::Receiver<ReadControl>,
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            Db,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
        )>,
        db: Db,
        update_sender: mpsc::Sender<UpdateMessage>,
        // jabber server must be able to both terminate the connection from error, and ask for data from the client (such as supported XEPs)
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        write_handle: WriteHandle,
        tasks: JoinSet<()>,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
    ) -> Self {
        let (_send, recv) = oneshot::channel();
        Self {
            control_receiver,
            stream,
            on_crash,
            db,
            update_sender,
            supervisor_control,
            write_handle,
            tasks,
            disconnecting: false,
            disconnect_timedout: recv,
            pending_iqs,
        }
    }

    async fn run(mut self) {
        println!("started read thread");
        // let stanza = self.stream.read::<Stanza>().await;
        // println!("{:?}", stanza);
        loop {
            tokio::select! {
                // if still haven't received the end tag in time, just kill itself
                // TODO: is this okay??? what if notification thread dies?
                Ok(()) = &mut self.disconnect_timedout => {
                    info!("disconnect_timedout");
                    break;
                }
                Some(msg) = self.control_receiver.recv() => {
                    match msg {
                        // when disconnect received,
                        ReadControl::Disconnect => {
                            let (send, recv) = oneshot::channel();
                            self.disconnect_timedout = recv;
                            self.disconnecting = true;
                            tokio::spawn(async {
                                tokio::time::sleep(Duration::from_secs(10)).await;
                                let _ = send.send(());
                            })
                        },
                        ReadControl::Abort(sender) => {
                            let _ = sender.send((self.db, self.update_sender, self.tasks, self.supervisor_control, self.write_handle, self.pending_iqs.clone()));
                            break;
                        },
                    };
                },
                s = self.stream.read::<Stanza>() => {
                    println!("read stanza");
                    match s {
                        Ok(s) => {
                            self.tasks.spawn(handle_stanza(s, self.update_sender.clone(), self.db.clone(), self.supervisor_control.clone(), self.write_handle.clone(), self.pending_iqs.clone()));
                        },
                        Err(e) => {
                            println!("error: {:?}", e);
                            // TODO: NEXT write the correct error stanza depending on error, decide whether to reconnect or properly disconnect, depending on if disconnecting is true
                            // match e {
                            //     peanuts::Error::ReadError(error) => todo!(),
                            //     peanuts::Error::Utf8Error(utf8_error) => todo!(),
                            //     peanuts::Error::ParseError(_) => todo!(),
                            //     peanuts::Error::EntityProcessError(_) => todo!(),
                            //     peanuts::Error::InvalidCharRef(_) => todo!(),
                            //     peanuts::Error::DuplicateNameSpaceDeclaration(namespace_declaration) => todo!(),
                            //     peanuts::Error::DuplicateAttribute(_) => todo!(),
                            //     peanuts::Error::UnqualifiedNamespace(_) => todo!(),
                            //     peanuts::Error::MismatchedEndTag(name, name1) => todo!(),
                            //     peanuts::Error::NotInElement(_) => todo!(),
                            //     peanuts::Error::ExtraData(_) => todo!(),
                            //     peanuts::Error::UndeclaredNamespace(_) => todo!(),
                            //     peanuts::Error::IncorrectName(name) => todo!(),
                            //     peanuts::Error::DeserializeError(_) => todo!(),
                            //     peanuts::Error::Deserialize(deserialize_error) => todo!(),
                            //     peanuts::Error::RootElementEnded => todo!(),
                            // }
                            // TODO: make sure this only happens when an end tag is received
                            if self.disconnecting == true {
                                break;
                            } else {
                                // AAAAAAAAAAAAAAAAAAAAA i should really just have this stored in the supervisor and not gaf bout passing these references around
                                let _ = self.on_crash.send((self.db, self.update_sender, self.tasks, self.supervisor_control, self.write_handle, self.pending_iqs.clone()));
                            }
                            break;
                        },
                    }
                },
                else => break
            }
        }
        println!("stopping read thread");
        // when it aborts, must clear iq map no matter what
        let mut iqs = self.pending_iqs.lock().await;
        for (_id, sender) in iqs.drain() {
            let _ = sender.send(Err(ReadError::LostConnection));
        }
    }
}

// what do stanza processes do?
// - update ui
// - access database
// - disconnect proper, reconnect
// - respond to server requests
async fn handle_stanza(
    stanza: Stanza,
    update_sender: mpsc::Sender<UpdateMessage>,
    db: Db,
    supervisor_control: mpsc::Sender<SupervisorCommand>,
    write_handle: WriteHandle,
    pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
) {
    match stanza {
        Stanza::Message(stanza_message) => {
            if let Some(mut from) = stanza_message.from {
                // TODO: group chat messages
                let mut message = Message {
                    id: stanza_message
                        .id
                        // TODO: proper id storage
                        .map(|id| Uuid::from_str(&id).unwrap_or_else(|_| Uuid::new_v4()))
                        .unwrap_or_else(|| Uuid::new_v4()),
                    from: from.clone(),
                    timestamp: Utc::now(),
                    body: Body {
                        // TODO: should this be an option?
                        body: stanza_message
                            .body
                            .map(|body| body.body)
                            .unwrap_or_default()
                            .unwrap_or_default(),
                    },
                };
                // TODO: can this be more efficient?
                let result = db
                    .create_message_with_user_resource_and_chat(message.clone(), from.clone())
                    .await;
                if let Err(e) = result {
                    tracing::error!("messagecreate");
                    let _ = update_sender
                        .send(UpdateMessage::Error(Error::MessageRecv(
                            MessageRecvError::MessageHistory(e.into()),
                        )))
                        .await;
                }
                message.from = message.from.as_bare();
                from = from.as_bare();
                let _ = update_sender
                    .send(UpdateMessage::Message { to: from, message })
                    .await;
            } else {
                let _ = update_sender
                    .send(UpdateMessage::Error(Error::MessageRecv(
                        MessageRecvError::MissingFrom,
                    )))
                    .await;
            }
        }
        Stanza::Presence(presence) => {
            if let Some(from) = presence.from {
                match presence.r#type {
                    Some(r#type) => match r#type {
                        // error processing a presence from somebody
                        stanza::client::presence::PresenceType::Error => {
                            // TODO: is there any other information that should go with the error? also MUST have an error, otherwise it's a different error. maybe it shoulnd't be an option.
                            let _ = update_sender
                                .send(UpdateMessage::Error(Error::Presence(
                                    // TODO: ughhhhhhhhhhhhh these stanza errors should probably just have an option, and custom display
                                    PresenceError::StanzaError(
                                        presence
                                            .errors
                                            .first()
                                            .cloned()
                                            .expect("error MUST have error"),
                                    ),
                                )))
                                .await;
                        }
                        // should not happen (error to server)
                        stanza::client::presence::PresenceType::Probe => {
                            // TODO: should probably write an error and restart stream
                            let _ = update_sender
                                .send(UpdateMessage::Error(Error::Presence(
                                    PresenceError::Unsupported,
                                )))
                                .await;
                        }
                        stanza::client::presence::PresenceType::Subscribe => {
                            // may get a subscription request from somebody who is not a contact!!! therefore should be its own kind of event
                            let _ = update_sender
                                .send(UpdateMessage::SubscriptionRequest(from))
                                .await;
                        }
                        stanza::client::presence::PresenceType::Unavailable => {
                            let offline = Offline {
                                status: presence.status.map(|status| status.status.0),
                            };
                            let _ = update_sender
                                .send(UpdateMessage::Presence {
                                    from,
                                    presence: Presence::Offline(offline),
                                })
                                .await;
                        }
                        // for now, do nothing, as these are simply informational. will receive roster push from the server regarding the changes to do with them.
                        stanza::client::presence::PresenceType::Subscribed => {}
                        stanza::client::presence::PresenceType::Unsubscribe => {}
                        stanza::client::presence::PresenceType::Unsubscribed => {}
                    },
                    None => {
                        let online = Online {
                            show: presence.show.map(|show| match show {
                                stanza::client::presence::Show::Away => Show::Away,
                                stanza::client::presence::Show::Chat => Show::Chat,
                                stanza::client::presence::Show::Dnd => Show::DoNotDisturb,
                                stanza::client::presence::Show::Xa => Show::ExtendedAway,
                            }),
                            status: presence.status.map(|status| status.status.0),
                            priority: presence.priority.map(|priority| priority.0),
                        };
                        let _ = update_sender
                            .send(UpdateMessage::Presence {
                                from,
                                presence: Presence::Online(online),
                            })
                            .await;
                    }
                }
            } else {
                let _ = update_sender
                    .send(UpdateMessage::Error(Error::Presence(
                        PresenceError::MissingFrom,
                    )))
                    .await;
            }
        }
        Stanza::Iq(iq) => match iq.r#type {
            stanza::client::iq::IqType::Error | stanza::client::iq::IqType::Result => {
                let send;
                {
                    send = pending_iqs.lock().await.remove(&iq.id);
                }
                if let Some(send) = send {
                    send.send(Ok(Stanza::Iq(iq)));
                } else {
                    let _ = update_sender
                        .send(UpdateMessage::Error(Error::Iq(IqError::NoMatchingId(
                            iq.id,
                        ))))
                        .await;
                }
            }
            // TODO: send unsupported to server
            // TODO: proper errors i am so tired please
            stanza::client::iq::IqType::Get => {}
            stanza::client::iq::IqType::Set => {
                if let Some(query) = iq.query {
                    match query {
                        stanza::client::iq::Query::Roster(mut query) => {
                            // TODO: there should only be one
                            if let Some(item) = query.items.pop() {
                                match item.subscription {
                                    Some(stanza::roster::Subscription::Remove) => {
                                        db.delete_contact(item.jid.clone()).await;
                                        update_sender
                                            .send(UpdateMessage::RosterDelete(item.jid))
                                            .await;
                                        // TODO: send result
                                    }
                                    _ => {
                                        let contact: Contact = item.into();
                                        if let Err(e) = db.upsert_contact(contact.clone()).await {
                                            let _ = update_sender
                                                .send(UpdateMessage::Error(Error::Roster(
                                                    RosterError::Cache(e.into()),
                                                )))
                                                .await;
                                        }
                                        let _ = update_sender
                                            .send(UpdateMessage::RosterUpdate(contact))
                                            .await;
                                        // TODO: send result
                                        // write_handle.write(Stanza::Iq(stanza::client::iq::Iq {
                                        //     from: ,
                                        //     id: todo!(),
                                        //     to: todo!(),
                                        //     r#type: todo!(),
                                        //     lang: todo!(),
                                        //     query: todo!(),
                                        //     errors: todo!(),
                                        // }));
                                    }
                                }
                            }
                        }
                        // TODO: send unsupported to server
                        _ => {}
                    }
                } else {
                    // TODO: send error (unsupported) to server
                }
            }
        },
        Stanza::Error(error) => {
            let _ = update_sender
                .send(UpdateMessage::Error(Error::Stream(error)))
                .await;
            // TODO: reconnect
        }
        Stanza::OtherContent(content) => {
            let _ = update_sender.send(UpdateMessage::Error(Error::UnrecognizedContent(content)));
            // TODO: send error to write_thread
        }
    }
}

pub enum ReadControl {
    Disconnect,
    Abort(
        oneshot::Sender<(
            Db,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
        )>,
    ),
}

pub struct ReadControlHandle {
    sender: mpsc::Sender<ReadControl>,
    pub(crate) handle: JoinHandle<()>,
}

impl Deref for ReadControlHandle {
    type Target = mpsc::Sender<ReadControl>;

    fn deref(&self) -> &Self::Target {
        &self.sender
    }
}

impl DerefMut for ReadControlHandle {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.sender
    }
}

impl ReadControlHandle {
    pub fn new(
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            Db,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
        )>,
        db: Db,
        sender: mpsc::Sender<UpdateMessage>,
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        jabber_write: WriteHandle,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = Read::new(
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            jabber_write,
            JoinSet::new(),
            pending_iqs,
        );
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }

    pub fn reconnect(
        stream: BoundJabberReader<Tls>,
        on_crash: oneshot::Sender<(
            Db,
            mpsc::Sender<UpdateMessage>,
            JoinSet<()>,
            mpsc::Sender<SupervisorCommand>,
            WriteHandle,
            Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
        )>,
        db: Db,
        sender: mpsc::Sender<UpdateMessage>,
        supervisor_control: mpsc::Sender<SupervisorCommand>,
        jabber_write: WriteHandle,
        tasks: JoinSet<()>,
        pending_iqs: Arc<Mutex<HashMap<String, oneshot::Sender<Result<Stanza, ReadError>>>>>,
    ) -> Self {
        let (control_sender, control_receiver) = mpsc::channel(20);

        let actor = Read::new(
            control_receiver,
            stream,
            on_crash,
            db,
            sender,
            supervisor_control,
            jabber_write,
            tasks,
            pending_iqs,
        );
        let handle = tokio::spawn(async move { actor.run().await });

        Self {
            sender: control_sender,
            handle,
        }
    }
}
