use std::sync::Arc;

use stanza::client::Stanza;
use thiserror::Error;
use tokio::{
    sync::{mpsc::error::SendError, oneshot::error::RecvError},
    time::error::Elapsed,
};

#[derive(Debug, Error, Clone)]
pub enum Error {
    #[error("already connected")]
    AlreadyConnected,
    // TODO: change to Connecting(ConnectingError)
    #[error("connecting: {0}")]
    Connecting(#[from] ConnectionError),
    #[error("presence: {0}")]
    Presence(#[from] PresenceError),
    #[error("set status: {0}")]
    SetStatus(#[from] StatusError),
    // TODO: have different ones for get/update/set
    #[error("roster: {0}")]
    Roster(RosterError),
    #[error("stream error: {0}")]
    Stream(#[from] stanza::stream::Error),
    #[error("message send error: {0}")]
    MessageSend(MessageSendError),
    #[error("message receive error: {0}")]
    MessageRecv(MessageRecvError),
    #[error("already disconnected")]
    AlreadyDisconnected,
    #[error("lost connection")]
    LostConnection,
    // TODO: Display for Content
    #[error("received unrecognized/unsupported content: {0:?}")]
    UnrecognizedContent(peanuts::element::Content),
    #[error("iq receive error: {0}")]
    Iq(IqError),
    #[error("disconnected")]
    Disconnected,
}

#[derive(Debug, Error, Clone)]
pub enum CommandError<T> {
    #[error("actor: {0}")]
    Actor(ActorError),
    #[error("{0}")]
    Error(#[from] T),
}

#[derive(Debug, Error, Clone)]
pub enum MessageSendError {
    #[error("could not add to message history: {0}")]
    MessageHistory(#[from] DatabaseError),
}

#[derive(Debug, Error, Clone)]
pub enum PresenceError {
    #[error("unsupported")]
    Unsupported,
    #[error("missing from")]
    MissingFrom,
    #[error("stanza error: {0}")]
    StanzaError(#[from] stanza::client::error::Error),
}

#[derive(Debug, Error, Clone)]
// TODO: should probably have all iq query related errors here, including read, write, stanza error, etc.
pub enum IqError {
    #[error("no iq with id matching `{0}`")]
    NoMatchingId(String),
}

#[derive(Debug, Error, Clone)]
pub enum MessageRecvError {
    #[error("could not add to message history: {0}")]
    MessageHistory(#[from] DatabaseError),
    #[error("missing from")]
    MissingFrom,
}

#[derive(Debug, Clone, Error)]
pub enum ConnectionError {
    #[error("connection failed: {0}")]
    ConnectionFailed(#[from] jabber::Error),
    #[error("failed roster retreival: {0}")]
    RosterRetreival(#[from] RosterError),
    #[error("failed to send available presence: {0}")]
    SendPresence(#[from] WriteError),
    #[error("cached status: {0}")]
    StatusCacheError(#[from] DatabaseError),
}

#[derive(Debug, Error, Clone)]
pub enum RosterError {
    #[error("cache: {0}")]
    Cache(#[from] DatabaseError),
    #[error("stream write: {0}")]
    Write(#[from] WriteError),
    // TODO: display for stanza, to show as xml, same for read error types.
    #[error("unexpected reply: {0:?}")]
    UnexpectedStanza(Stanza),
    #[error("stream read: {0}")]
    Read(#[from] ReadError),
    #[error("stanza error: {0}")]
    StanzaError(#[from] stanza::client::error::Error),
}

#[derive(Debug, Error, Clone)]
#[error("database error: {0}")]
pub struct DatabaseError(Arc<sqlx::Error>);

#[derive(Debug, Error, Clone)]
pub enum DatabaseOpenError {
    #[error("error: {0}")]
    Error(Arc<sqlx::Error>),
    #[error("migration: {0}")]
    Migration(Arc<sqlx::migrate::MigrateError>),
    #[error("io: {0}")]
    Io(Arc<tokio::io::Error>),
    #[error("invalid path")]
    InvalidPath,
}

impl From<sqlx::Error> for DatabaseError {
    fn from(e: sqlx::Error) -> Self {
        Self(Arc::new(e))
    }
}

impl From<sqlx::Error> for DatabaseOpenError {
    fn from(e: sqlx::Error) -> Self {
        Self::Error(Arc::new(e))
    }
}

impl From<sqlx::migrate::MigrateError> for DatabaseOpenError {
    fn from(e: sqlx::migrate::MigrateError) -> Self {
        Self::Migration(Arc::new(e))
    }
}

impl From<tokio::io::Error> for DatabaseOpenError {
    fn from(e: tokio::io::Error) -> Self {
        Self::Io(Arc::new(e))
    }
}

#[derive(Debug, Error, Clone)]
pub enum StatusError {
    #[error("cache: {0}")]
    Cache(#[from] DatabaseError),
    #[error("stream write: {0}")]
    Write(#[from] WriteError),
}

#[derive(Debug, Error, Clone)]
pub enum WriteError {
    #[error("xml: {0}")]
    XML(#[from] peanuts::Error),
    #[error("lost connection")]
    LostConnection,
    // TODO: should this be in writeerror or separate?
    #[error("actor: {0}")]
    Actor(#[from] ActorError),
    #[error("disconnected")]
    Disconnected,
}

// TODO: separate peanuts read and write error?
#[derive(Debug, Error, Clone)]
pub enum ReadError {
    #[error("xml: {0}")]
    XML(#[from] peanuts::Error),
    #[error("lost connection")]
    LostConnection,
}

#[derive(Debug, Error, Clone)]
pub enum ActorError {
    #[error("receive timed out")]
    Timeout,
    #[error("could not send message to actor, channel closed")]
    Send,
    #[error("could not receive message from actor, channel closed")]
    Receive,
}

impl From<Elapsed> for ActorError {
    fn from(_e: Elapsed) -> Self {
        Self::Timeout
    }
}

impl<T> From<SendError<T>> for ActorError {
    fn from(_e: SendError<T>) -> Self {
        Self::Send
    }
}

impl From<RecvError> for ActorError {
    fn from(_e: RecvError) -> Self {
        Self::Receive
    }
}
